"""Plugins for Regular Expected Transactions."""

from beancount.core.data import Entries, Options, Transaction
from beanahead.utils import TAG_RX, RX_META_DFLTS

__plugins__ = ["convert_meta"]


def convert_meta(entries: Entries, _: Options) -> tuple[Entries, list[tuple]]:
    """Set default values for Regular Expected Transaction meta fields.

    Parameters
    ----------
    entries
        Entries being loaded.
    """
    for entry in entries:
        if (not isinstance(entry, Transaction)) or TAG_RX not in entry.tags:
            continue
        for k, v in RX_META_DFLTS.items():
            entry.meta.setdefault(k, v)

    return entries, (errors := [])



"""Functions to administer expired transactions."""

from __future__ import annotations

import datetime
from datetime import timedelta
from pathlib import Path
import re

from beancount.core.data import Transaction

from . import utils
from .errors import BeanaheadWriteError

TODAY = utils.TODAY
TOMORROW = TODAY + timedelta(1)

DATE_FORMATS = "(YYYY-MM-DD or MM-DD or DD)"
VALID_DATE_FORMAT_REGEXES = [
    re.compile(r"^\d{4}-\d{1,2}-\d{1,2}$"),
    re.compile(r"^\d{1,2}-\d{1,2}$"),
    re.compile(r"^\d{1,2}$"),
]


def is_valid_date_format(string: str) -> bool:
    """Query if a string is in valid date format.

    Examples
    --------
    >>> is_valid_date_format("2022-08-15")
    True
    >>> is_valid_date_format("2022-8-15")
    True
    >>> is_valid_date_format("08-15")
    True
    >>> is_valid_date_format("15-08")  # NB Not a valid date, but a valid date format
    True
    >>> is_valid_date_format("15")
    True
    >>> is_valid_date_format("22-08-15")
    False
    >>> is_valid_date_format("2022")
    False
    >>> is_valid_date_format("08-15-2022")
    False
    >>> is_valid_date_format("08-15-22")
    False
    """
    return any(regex.match(string) is not None for regex in VALID_DATE_FORMAT_REGEXES)


def _get_date_parts(response: str) -> tuple[int, int, int]:
    """Return parts of date to be formed from user response.

    If response does not define a month then month will be returned
    as the current month or the next month depending on whether the
    defined day is, respectively, >= or < the current day.

    If response does not define a year then year will be returned as
    the current year or the next year, depending on whether the defined
    or eveluated month/day combination represent a day, respectively,
    >= or < the current day.

    Parameters
    ----------
    response
        User response to request for date. `response` must match
        a regex of `valid_date_format_regexes`
    """
    parts = [int(part) for part in response.split("-")]
    day = parts[-1]
    if len(parts) == 3:
        year, month = parts[0:2]
    else:
        if len(parts) == 2:
            month = parts[0]
        elif day >= TODAY.day:
            month = TODAY.month
        else:
            month = 1 if TODAY.month + 1 == 13 else TODAY.month + 1

        if month == TODAY.month:
            year = TODAY.year if day >= TODAY.day else TODAY.year + 1
        else:
            year = TODAY.year if month > TODAY.month else TODAY.year + 1
    return year, month, day


def _get_date_from_response(response: str) -> datetime.date | None:
    """Get date corresponding with user input.

    Returns
    -------
    datetime.date or None
        datetime.date representing `response`.
        None if `response` does not represent a valid date.
    """
    if not is_valid_date_format(response):
        return None
    parts = _get_date_parts(response)
    try:
        return datetime.date(*parts)
    except ValueError:
        return None


def _update_txn(txn: Transaction, path: Path) -> Transaction | None:
    """Return updated expired transaction according to user response.

    Parameters
    ----------
    txn
        Expired expected transaction to update.

    path
        Path to ledger on which txn defined.

    Returns
    -------
    Transaction
        Existing `txn` if user chooses to leave 'as is'.
        None if user choose to remove.
        Otherwise a new Transaction object with revised date.
    """
    utils.print_it(
        f"{utils.SEPARATOR_LINE}\nThe following transaction has expired."
        f"\n\n{utils.compose_entries_content(txn)}"
        f"\n0 Move transaction forwards to tomorrow ({TOMORROW})."
        f"\n1 Move transaction forwards to another date."
        f"\n2 Remove transaction from ledger {path.stem}."
        f"\n3 Leave transaction as is."
    )
    response: str = utils.get_input("Choose one of the above options, [0-3]:")
    while not utils.response_is_valid_number(response, 3):
        response = utils.get_input(
            f"'{response}' is not valid input, please try again, [0-3]: "
        )
    if response == "3":
        return txn
    elif response == "2":
        return None
    elif response[0] == "0":
        return txn._replace(date=TOMORROW)

    response = utils.get_input(f"Enter a new date {DATE_FORMATS}: ")
    while (date := _get_date_from_response(response)) is None or (date < TODAY):
        if date is None:
            response = utils.get_input(
                f"'{response}' does not represent a valid date."
                f"\nPlease enter a valid date {DATE_FORMATS}: "
            )
        else:
            response = utils.get_input(
                f"'{response}' does not represent a date >= {TODAY}."
                f"\nPlease enter a valid date {DATE_FORMATS}: "
            )
    return txn._replace(date=date)


def overwrite_ledgers(contents: dict[Path, str]):
    """Write content to expected transaction ledgers.

    If an error is raised during the writing process then:
        Attempts to revert all overwritten ledgers to prior content.

        Raises BeanaheadWriteError advising which ledgers were
        reverted.

    Parameters
    ----------
    contents : dict
        key : Path
            Path of ledger to which to write 'value'.
        value: str
            Content to write to 'key'.

    Raises
    ------
    BeanaheadWriteError
        If any error is raised during any write process.
    """
    prev_contents = {path: utils.get_content(path) for path in contents}
    seen = []
    try:
        for path, content in contents.items():
            seen.append(path)
            utils.overwrite_file(path, content)
    except Exception as err:
        for path_ in seen[:]:
            try:
                utils.write(path_, prev_contents[path_])
            except Exception:
                seen.remove(path_)
        raise BeanaheadWriteError(path, seen) from err


def admin_expired_txns(ledgers: list[str]):
    """Administer expired expected transactions.

    For each expired transaction on a ledger of `ledgers`:
        Requests user choose from following options:
            Move txn forwards to tomorrow.
            Move txn forwards to user-defined date.
            Remove transaction from ledger.
            Leave transaction as is.

        Actions request.

    Ledgers on which at least one transaction is moved forward or removed
    will be rewritten. As part of this all remaining entries are sorted
    in ascending order.

    Parameters
    ----------
    ledgers
        List of paths to expected transactions ledgers. Paths can
        be defined as absolute or relative to the cwd. It is not
        necessary to include the. beancount extension. For example,
        "rx" would refer to the file 'rx.beancount' in the cwd.
    """
    x_txns: dict[Path, list[Transaction]] = {}
    file_keys: dict[Path, str] = {}
    for ledger in ledgers:
        path = utils.get_verified_path(ledger)
        file_keys[path] = utils.get_verified_ledger_file_key(path)
        x_txns[path] = utils.get_unverified_txns(path)

    no_expired_txns = True
    paths = list(x_txns.keys())
    ledger_updated = {path: False for path in paths}
    updated_txns: dict[path, list[Transaction]] = {}
    for path, txns in x_txns.items():
        new_txns = []
        for txn in txns:
            if utils.is_expired(txn):
                no_expired_txns = False
                txn_ = _update_txn(txn, path)
                if not ledger_updated[path] and txn_ != txn:
                    ledger_updated[path] = True
                txn = txn_
            if txn is not None:
                new_txns.append(txn)
        updated_txns[path] = new_txns

    if no_expired_txns:
        paths_string = "\n".join([str(path) for path in paths])
        utils.print_it(
            "There are no expired transactions on any of the following"
            f" ledgers:\n{paths_string}"
        )
        return

    updated_paths = [path for path in paths if ledger_updated[path]]
    paths_string = "\n".join([str(path) for path in updated_paths])
    if not updated_paths:
        utils.print_it(
            "\nYou have not choosen to modify any expired transactions."
            "\nNo ledger has been altered."
        )
        return

    updated_contents: dict[Path, str] = {}
    for path in updated_paths:
        content = utils.create_ledger_content(file_keys[path], updated_txns[path])
        updated_contents[path] = content

    overwrite_ledgers(updated_contents)
    utils.print_it(f"\nThe following ledgers have been updated:\n{paths_string}")



"""beanahead pacakge constructor."""

from importlib.metadata import version
from pathlib import Path
import sys

from . import plugins

__copyright__ = "Copyright (c) 2022 Marcus Read"

# insert plugins directory to path.
path = Path(plugins.__file__).parent
sys.path.insert(0, str(path))


# Resolve version
__version__ = None

try:
    # get version from installed package
    __version__ = version("beanahead")
except ImportError:
    pass

if __version__ is None:
    try:
        # if package not installed, get version as set when package built
        from ._version import version
    except Exception:
        # If package not installed and not built, leave __version__ as None
        pass
    else:
        __version__ = version

del version



"""Functions to reconcile new transactions with expected transactions."""

from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable
from datetime import timedelta
from decimal import Decimal
from pathlib import Path
import re

from beancount.core import data
from beancount.core.data import Transaction
from beancount.core.getters import get_entry_accounts
from beancount.core import number
from beancount.parser.parser import parse_file
from beangulp.extract import HEADER

from . import utils
from .errors import BeanaheadWriteError


def separate_out_txns(entries: data.Entries) -> tuple[list[Transaction], data.Entries]:
    """Separate transactions from other entries."""
    txns, others = [], []
    for entry in entries:
        lst = txns if isinstance(entry, Transaction) else others
        lst.append(entry)
    return txns, others


def get_close_txns(
    txns: list[Transaction], x_txn: Transaction, delta: timedelta
) -> list[Transaction]:
    """Return txns that are close to the date of an Expected Transaction.

    Parameters
    ----------
    txns
        Transactions from which to return those that are within `delta`
        from the date of `x_txn`.

    x_txn
        Expected Transaction to query.

    delta
        Delta, in days, that defines how close the date of a txn (of
        `txns`) should be to the date of the `x_txn` in order to be
        included in return.
    """
    start, end = x_txn.date - delta, x_txn.date + delta + timedelta(1)
    return list(data.iter_entry_dates(txns, start, end))


def bal_sheet_accounts_match(a: Transaction, b: Transaction) -> bool:
    """Query if two transactions have postings to same bal sheet acc."""
    accounts_a = utils.get_balance_sheet_accounts(a)
    accounts_b = utils.get_balance_sheet_accounts(b)
    return bool(set(accounts_a).intersection(set(accounts_b)))


def get_accounts_matches(
    txns: list[Transaction], x_txn: Transaction
) -> list[Transaction]:
    """Return txns matching an Expected Transaction's bal sheet acc.

    Returns
    -------
    list of Transaction
        Transactions of `txns` with a posting to a balance sheet account
        to which the `x_txn` also has a posting.
    """
    return [txn for txn in txns if bal_sheet_accounts_match(txn, x_txn)]


def get_basic_matches(txns: list[Transaction], x_txn: Transaction) -> list[Transaction]:
    """Return transactions that basically match an Expected Transaction.

    Returns
    -------
    list of Transaction
        Transactions of `txns` that match `x_txn` based on:
            Closeness of dates (5 days)
            Postings to same balance sheet account.
    """
    matches = get_close_txns(txns, x_txn, timedelta(days=5))
    return get_accounts_matches(matches, x_txn)


def get_pattern(x_txn: Transaction) -> re.Pattern:
    """Get a pattern to match an Expected Transaction's payee"""
    payee = x_txn.payee
    words = payee.split(" ")
    if len(words) == 1:
        return re.compile(payee, re.I)
    return utils.compile_strings_regex(words)


def get_payee_matches(txns: list[Transaction], x_txn: Transaction) -> list[Transaction]:
    """Return transactions matching an Expected Transaction's payee."""
    pattern = get_pattern(x_txn)
    return [
        txn for txn in txns if (txn.payee and pattern.search(txn.payee) is not None)
    ]


def get_common_accounts(a: Transaction, b: Transaction) -> set[str]:
    """Return set of accounts that are common to two transactions."""
    return get_entry_accounts(a).intersection(get_entry_accounts(b))


def get_common_balance_sheet_accounts(a: Transaction, b: Transaction) -> set[str]:
    """Return set of balance sheet accounts common to two transactions."""
    common_accounts = get_common_accounts(a, b)
    return {acc for acc in common_accounts if utils.is_balance_sheet_account(acc)}


def get_posting_to_account(txn: Transaction, account: str) -> data.Posting | None:
    """Return a transaction's posting to a given account

    None if transaction does not have a posting to `account`.
    """
    postings = [post for post in txn.postings if post.account == account]
    if len(postings) > 1:
        raise ValueError(
            "Transaction cannot have multiple postings to the same account"
            " although the following transaction has multiple postings to"
            f" '{account}':\n{txn}"
        )
    if not postings:
        return None
    return postings[0]


def get_amount_for_account(txn: Transaction, account: str) -> data.Amount | None:
    """Return amount of a transaction's posting to a given account.

    Returns None if no posting to `account`.
    """
    posting = get_posting_to_account(txn, account)
    return None if posting is None else posting.units


def decimal_diff(a: Decimal, b: Decimal) -> Decimal:
    """Return difference of a decimal relative to another.

    Returns
    -------
    Decimal
        Decimal bewteen 0 and 1 representing percentage difference of
        larger number relative to smaller number.

        Decimal(1) if the signs of `a` and `b` are different.

    Examples
    --------
    >>> decimal_diff(Decimal("40"), Decimal("40"))
    Decimal('0')
    >>> decimal_diff(Decimal("40"), Decimal("50"))
    Decimal('0.2')
    >>> decimal_diff(Decimal("50"), Decimal("40"))
    Decimal('0.2')
    >>> decimal_diff(Decimal("-50"), Decimal("-40"))
    Decimal('0.2')
    >>> decimal_diff(Decimal("-50"), Decimal("40"))
    Decimal('1')
    """
    if a * b < number.ZERO:
        return number.ONE
    denom = max(abs(a), abs(b))
    return abs(a - b) / denom


def number_diff(a: Transaction, b: Transaction) -> Decimal:
    """Return percentage difference between number of two transactions.

    Return represents percentage difference of number(s) of `b` from `a`.

    Returns
    -------
    Decimal
        Decimal bewteen 0 and 1 representing average percentage difference
        between transactions' postings' 'number' (only for postings
        to balance sheet accounts that are common to both transactions).
        Decimal(0) indicates the number of postings' to all common balance
        sheet accounts are exactly the same.

        Returns Decimal(1) if
            There are no common balance sheet accounts with number
            denominated in same currency.

            For a common balance sheet account the sign of the number on
            each transaction is different.
    """
    diffs = []
    for account in get_common_balance_sheet_accounts(a, b):
        num_a, cur_a = get_amount_for_account(a, account)
        num_b, cur_b = get_amount_for_account(b, account)
        if not cur_a == cur_b:
            continue
        diff = decimal_diff(num_a, num_b)
        if diff == number.ONE:
            return diff
        diffs.append(diff)
    if not diffs:
        return number.ONE
    return abs(sum(diffs) / len(diffs))


def get_sortkey_number(x_txn: Transaction) -> Callable:
    """Get sortkey function to sort transactions by difference in number.

    Sortkey function will sort transactions by difference in number to
    `x_txn`.
    """

    def sortkey_number(txn: Transaction) -> Decimal:
        """sortkey to sort transactions by difference in number."""
        return number_diff(x_txn, txn)

    return sortkey_number


def sort_by_number(txns: list[Transaction], x_txn: Transaction) -> list[Transaction]:
    """Sort transactions by difference in number with Expected Transaction.

    Returns sorted copy of transactions.
    """
    return sorted(txns, key=get_sortkey_number(x_txn))


def get_sortkey_date(x_txn: Transaction) -> Callable:
    """Get sortkey function to sort transactions by difference in dates.

    Sortkey function will sort transactions by, in order of priority:
        difference in date with `x_txn`.
        number of common accounts.
    """

    def sortkey_date(txn: Transaction) -> Decimal:
        """sortkey to sort transactions by difference in date."""
        return (abs(x_txn.date - txn.date), -len(get_common_accounts(txn, x_txn)))

    return sortkey_date


def sort_by_date(txns: list[Transaction], x_txn: Transaction) -> list[Transaction]:
    """Sort transactions by difference in date with an expected transaction.

    Returns sorted copy of transactions.
    """
    return sorted(txns, key=get_sortkey_date(x_txn))


def have_same_number(a: Transaction, b: Transaction) -> bool:
    """Query if two transactions' postings to common accounts have same number."""
    return number_diff(a, b) == number.ZERO


def get_number_matches(
    txns: list[Transaction], x_txn: Transaction, margin: Decimal = number.ZERO
) -> list[Transaction]:
    """Return transactions matching expected transaction's numbers."""
    return [txn for txn in txns if number_diff(x_txn, txn) <= margin]


def get_matches(txns: list[Transaction], x_txn: Transaction) -> list[Transaction]:
    """Match an Expected Transaction to one or more incoming transactions."""
    basic_matches = get_basic_matches(txns, x_txn)
    if not basic_matches:
        return []

    payee_matches = get_payee_matches(basic_matches, x_txn)
    if not payee_matches:
        # if no matches by payee then return only those within 2%
        matches = get_number_matches(basic_matches, x_txn, Decimal("0.02"))
        return sort_by_date(matches, x_txn)
    if len(payee_matches) == 1:
        return payee_matches

    matches = sort_by_date(payee_matches, x_txn)
    closest_match = matches[0]
    # if closest match has exact value, return only those with that date and exact value
    if have_same_number(closest_match, x_txn):
        delta = abs(closest_match.date - x_txn.date)
        return [
            txn
            for txn in matches
            if abs(txn.date - x_txn.date) == delta and have_same_number(txn, x_txn)
        ]
    return sort_by_number(matches, x_txn)


MSG_SINGLE_MATCH = "Do you want to match the above transactions? y/n: "


def confirm_single(
    x_txn: Transaction, matches: list[Transaction]
) -> Transaction | None:
    """Get user to decide if two transactions should be matched.

    Parameters
    ----------
    x_txn
        Expected transaction

    matches
        List of length 1 containing incoming match.

    Returns
    -------
    pd.Transaction or None
        Matched transaction, if user confirm match.
        None if user rejects match.
    """
    utils.print_it(
        f"{utils.SEPARATOR_LINE}Expected Transaction:\n"
        f"{utils.compose_entries_content(x_txn)}\n"
        f"Incoming Transaction:\n{utils.compose_entries_content(matches[0])}"
    )
    response = utils.get_input(MSG_SINGLE_MATCH).lower()
    while response not in ["n", "y"]:
        response = utils.get_input(
            f"{response} is not valid input, please try again, y/n: "
        )
    if response == "n":
        return None
    elif response == "y":
        return matches[0]


def get_mult_match(
    x_txn: Transaction, matches: list[Transaction]
) -> Transaction | None:
    """Get user to choose a match from various options.

    Requests user decides which of `matches` to match with `x_txn`, or to
    reject all matches.

    Returns
    -------
    pd.Transaction or None
        Matched transaction, as choosen by user.
        None if user rejects all matches.
    """
    utils.print_it(
        f"{utils.SEPARATOR_LINE}Expected Transaction:\n"
        f"{utils.compose_entries_content(x_txn)}\n\n"
        f"Incoming Transactions:\n"
    )
    for i, match in enumerate(matches):
        utils.print_it(f"{i}\n{utils.compose_entries_content(match)}")

    max_value = len(matches) - 1
    options = f"[0-{max_value}]/n"
    response = utils.get_input(
        "Which of the above incoming transactions do you wish to match"
        f" with the expected transaction, or 'n' for None, {options}:"
    )
    while not (
        (response == "n") or utils.response_is_valid_number(response, max_value)
    ):
        response = utils.get_input(
            f"{response} is not valid input, please try again {options}: "
        )
    if response == "n":
        return None
    return matches[int(response)]


def reconcile_x_txns(
    x_txns: list[Transaction], new_txns: list[Transaction]
) -> list[tuple[Transaction, Transaction]]:
    """Reconcile Expected Transactions with new transactions.

    Requests user confirm or decide matches.

    Parameters
    ----------
    x_txns
        Expected Transactions to be reconciled.

    new_txns
        Incoming transactions to reconile Expected Transactions against.

    Returns
    -------
    list of 2-tuple of Transaction
        Each 2-tuple represents a match, where
            [0] Expected Transaction
            [1] new transaction matched with [0]
        Only matched Expected Transactions, of `x_txns`, are inlcuded in
        return.
    """
    reconciled = []
    new_txns = new_txns.copy()
    for x_txn in x_txns:
        matches = get_matches(new_txns, x_txn)
        if not matches:
            continue
        match_func = confirm_single if len(matches) == 1 else get_mult_match
        match = match_func(x_txn, matches)
        if match is not None:
            reconciled.append((x_txn, match))
            new_txns.remove(match)
    return reconciled


def update_new_txn(new_txn: Transaction, x_txn: Transaction) -> Transaction:
    """Update a incoming txn with data from a matched Expected Transaction.

    The following fields of the `new_txn` will be updated with the
    corresponding fields of `x_txn`:
        narration
        tags (excluding expected transaction tags, for example #rx_txn)
        meta:
            Items of `x_txn` meta will be added to the `new_txn` meta if
            the key is not already present. This excludes beanahead meta
            fields ('final', 'roll' etc) which will not be added to the
            `new_txn`.
        postings:
            postings to the `x_txn` will be added to the `new_txn` posting
            if the `new_txn` does not otherwise include a posting to the
            corresponding account.

            If the `new_txn` and `x_txn` include postings to the same
            account and only the `x_txn` defines a number, the posting to
            the `new_txn` will be updated to reflect the value as defined
            on the `x_txn`

    All other fields ('date', 'payee' etc), will remain as defined on the
    `new_txn`.

    Parameters
    ----------
    new_txn
        Incoming transaction matched with `x_txn`.

    x_txn
        Expected transaction matched with `new_txn`.

    Returns
    -------
    Transaction
        Copy of `new_txn` updated to reflect any complementary data
        provided by `x_txn`.

    Notes
    -----
    For the purpose of evaluating which postings should be updated,
    postings are treated as defined on underlying ledger from which data
    loaded - i.e. any numbers that are interpolated on loading are ignored.
    """
    # reverse_automatic_balancing returns copies
    x_txn = utils.reverse_automatic_balancing(x_txn)
    new_txn = utils.reverse_automatic_balancing(new_txn)

    if not new_txn.narration:
        new_txn = new_txn._replace(narration=x_txn.narration)

    if tags_to_add := x_txn.tags - utils.TAGS_X:
        new_txn = utils.add_tags(new_txn, tags_to_add)

    new_txn = new_txn._replace(meta=new_txn.meta.copy())
    for k, v in x_txn.meta.items():
        if k in utils.RX_META_DFLTS or k in new_txn.meta or k == "freq":
            continue
        new_txn.meta[k] = v

    new_txn_accounts = get_entry_accounts(new_txn)
    for posting in x_txn.postings:
        if (account := posting.account) not in new_txn_accounts:
            new_txn.postings.append(posting)
        else:
            new_txn_posting = get_posting_to_account(new_txn, account)

            # carry over any meta not otherwise defined on new_txn
            if new_txn_posting.meta:
                updated_posting = new_txn_posting._replace(
                    meta=new_txn_posting.meta.copy()
                )
            else:
                updated_posting = new_txn_posting._replace(meta={})
            for k, v in posting.meta.items():
                updated_posting.meta.setdefault(k, v)

            if updated_posting.units is None:
                updated_posting = updated_posting._replace(units=posting.units)

            new_txn.postings.remove(new_txn_posting)
            new_txn.postings.append(updated_posting)
    return new_txn


def update_new_txns(
    new_txns: list[Transaction],
    reconciled_x_txns: list[tuple[Transaction, Transaction]],
) -> list[Transaction]:
    """Update incoming transactions for matched data.

    Parameters
    ----------
    new_txns
        Incoming transactions (from injection file).

    reconciled_x_txns : list of 2-tuple of Transaction
        Where each 2-tuple represents a matched expected transaction:
            [0] expected transaction
            [1] new txn, of `new_txns`, matched with [0]

    Returns
    -------
    list of Transaction
        `new_txns` as updated. Received order is NOT maintained.
    """
    for x_txn, new_txn in reconciled_x_txns:
        updated_txn = update_new_txn(new_txn, x_txn)
        new_txns.insert(0, updated_txn)
        new_txns.remove(new_txn)
    return new_txns


def map_path_to_reconciled_x_txns(
    x_txns: dict[Path, list[Transaction]],
    reconciled_x_txns: tuple[Transaction, Transaction],
) -> dict[Path, list[Transaction]]:
    """Map reconciled x_txns to ledgers they were defined on.

    Parameters
    ----------
    x_txns : dict
        key : Path
            Path to ledger on which 'value' defined.
        value : list of Transaction
            Expected Transactions defined on 'key'.

    reconciled_x_txns : 2-tuple of Transaction
        As return from reconcile_x_txns() function.

    Returns
    -------
    dict
        key: Path
            Path to ledger on which 'value' defined.
        value: list of Transaction
            Reconciled Expected Transactions defined on 'key'.
    """
    mapping = defaultdict(list)
    for x_txn, _ in reconciled_x_txns:
        mapped = False
        for path, txns in x_txns.items():
            if x_txn in txns:
                mapping[path].append(x_txn)
                mapped = True
                break
        assert mapped, "reconciled x_txn not mapped to path."
    return mapping


def reconcile_new_txns(
    new_entries: str,
    x_txns_ledgers: list[str],
    remove: bool = True,
    output: str | None = None,
    ascending: bool = True,
):
    """Reconcile new transactions with expected transactions.

        Evaluates potential matches between incoming and expected
        transactions.

        Requests user intervention to confirm potential matches
        and resolve ambiguities.

        Updates each matched incoming transaction with the following
        fields of the matched expected transaction:
            narration
            postings:
                postings to the x_txn will be added to the new_txn
                posting if the new_txn does not otherwise include a
                posting to the corresponding account.

                If the new_txn and x_txn include postings to the same
                account and only the x_txn defines a number, the posting
                to the 'new_txn' will be updated to reflect the value as
                defined on the 'x_txn'.

        Writes updated incoming transactions to output (by default
        overwrites `new_entries`).

        Optinally, removes matched expected transactions from the
        corrsponding expected transaction ledger by way of rewritting
        these ledgers with the remaining, unmatched, transactions.

    Parameters
    ----------
    Where a parameter takes one or more file addresses, the address can
    be defined as either absolute paths or paths relative to the cwd. It
    is not necessary to include the. beancount extension. For example,
    "rx" would refer to the file 'rx.beancount' in the cwd.

    new_entires : str
        Path to .beancount file containing incoming transactions and
        other new entries. Entries in file should represent ONLY new
        entries.

    x_txns_ledgers : list of str
        List of paths to expected transactions ledgers which list
        transactions that include transactions that may reconcile
        with one or more transactions of `new_txns`.

    remove : bool, default: True
        Should matched expected transactions be removed from the
        corresponding ledger of `x_txns_ledgers`. This has the effect
        rewritting the corresponding expected transaction ledger(s) with
        the remaining, unmatched, expected transactions. As part of this
        rewrite all entries are resorted in ascending order.

    output : str, default: overwrite file represented by `new_entries`
        Path to .beancount file to which the reconciled new entries
        should be written. Any existing file will be overwritten.

    ascending : bool, default: True
        True to order output earliest transfer first.
        False to order output latest transfer first.
    """
    input_path = utils.get_verified_path(new_entries)
    new_entries_, _, _ = parse_file(str(input_path))
    new_txns, new_other = separate_out_txns(new_entries_)

    x_txns: dict[Path, list[Transaction]] = {}
    for ledger in x_txns_ledgers:
        path = utils.get_verified_path(ledger)
        _ = utils.get_verified_ledger_file_key(path)  # just verify that a ledger
        x_txns[path] = utils.get_unverified_txns(path)

    all_x_txns = []
    for txns in x_txns.values():
        all_x_txns.extend(txns)

    reconciled_x_txns = reconcile_x_txns(all_x_txns, new_txns)
    updated_new_txns = update_new_txns(new_txns, reconciled_x_txns)
    updated_entries = updated_new_txns + new_other
    updated_entries.sort(key=data.entry_sortkey, reverse=not ascending)

    out_path = input_path if output is None else utils.get_unverified_path(output)
    out_content = HEADER + "\n" + utils.compose_entries_content(updated_entries)

    x_txns_to_remove = (
        map_path_to_reconciled_x_txns(x_txns, reconciled_x_txns) if remove else {}
    )

    # Write / Overwrite files, reverting all to previous content on any errors
    prev_contents = {path: utils.get_content(path) for path in x_txns_to_remove}
    seen: list[Path] = []
    try:
        for path, txns in x_txns_to_remove.items():
            seen.append(path)
            utils.remove_txns_from_ledger(path, txns)
    except Exception as err:
        for revert_path in seen[:]:
            try:
                utils.write(revert_path, prev_contents[revert_path])
            except Exception:
                seen.remove(revert_path)
        raise BeanaheadWriteError(path, seen) from err

    overwrite = False
    if out_path.is_file():
        prev_contents[out_path] = utils.get_content(out_path)
        overwrite = True
    try:
        utils.write(out_path, out_content)
    except Exception as err:
        reverted = list(prev_contents.keys())
        for revert_path, prev_content in prev_contents.items():
            try:
                utils.write(revert_path, prev_content)
            except Exception:
                reverted.remove(revert_path)
        raise BeanaheadWriteError(out_path, reverted, overwrite) from err

    msg = (
        f"{len(reconciled_x_txns)} incoming transactions have been reconciled against"
        f" expected transactions.\nUpdated transactions have been output to"
        f" '{out_path}'."
    )
    for path, txns in x_txns_to_remove.items():
        msg += f"\n{len(txns)} transactions have been removed from ledger {path}."

    utils.print_it(msg)



"""General utility functions."""

from __future__ import annotations

import copy
import datetime
from pathlib import Path
import re

from beancount import loader
from beancount.core import data
from beancount.core.account_types import is_account_type
from beancount.core.data import Transaction
from beancount.core.interpolate import AUTOMATIC_META
from beangulp.extract import HEADER
from beancount.parser import parser, printer

from . import config
from .config import get_account_root_names, BC_DEFAULT_ACCOUNT_ROOT_NAMES
from .errors import (
    BeancountFileExistsError,
    BeanaheadFileExistsError,
    BeanaheadLedgerFileExistsError,
    BeanaheadFileKeyError,
    BeancountLoaderErrors,
)

TAG_X = "x_txn"
TAG_RX = "rx_txn"
TAGS_X = set([TAG_X, TAG_RX])

RX_META_DFLTS = {
    "final": None,
    "roll": True,
}

SEPARATOR_LINE = "-" * 77 + "\n"
TODAY = datetime.datetime.now().date()

FILE_CONFIG = {
    "x": {
        "title": "Expected Transactions Ledger",
        "plugin": None,
        "tag": TAG_X,
        "comment": "Enter expected transactions after this line...",
        "post_comment": "...enter expected transactions before this line.",
    },
    "rx": {
        "title": "Regular Expected Transactions Ledger",
        "plugin": "rx_txn_plugin",
        "tag": TAG_RX,
        "comment": "Transactions should not be manually added to this file.",
        "post_comment": None,
    },
    "rx_def": {
        "title": "Regular Expected Transaction Definitions",
        "plugin": "rx_txn_plugin",
        "tag": TAG_RX,
        "comment": (
            "All accounts referenced by definitions should be defined on the main"
            " ledger.\nEnter definitions after this line..."
        ),
        "post_comment": "...enter definitions before this line.",
    },
}

LEDGER_FILE_KEYS = ["x", "rx"]


def print_it(text: str, **kwargs):
    """Print to the selected stream.

    Parameters
    ----------
    text
        Text to print.

    kwargs
        Kwargs to pass to 'print' function.
    """
    kwargs["file"] = config.SETTINGS.print_to
    print(text, **kwargs)


def validate_file_key(file_key: str):
    """Validate a file_key.

    Raises
    ------
    ValueError
        If `file_key` is not a valid key.
    """
    if file_key not in (valid_keys := FILE_CONFIG.keys()):
        raise ValueError(
            f"'{file_key}' is not a valid file key. `file_key` must be a"
            f" value in {list(valid_keys)}."
        )


def validate_ledger_file_key(file_key: str):
    """Validate a ledger file_key.

    Raises
    ------
    ValueError
        If `file_key` is not a valid ledger file key.
    """
    if file_key not in LEDGER_FILE_KEYS:
        raise ValueError(
            f"'{file_key}' is not a valid ledger file key. `file_key` must be"
            f" a value in {LEDGER_FILE_KEYS}."
        )


def compose_header_footer(file_key: str) -> tuple[str, str]:
    """Compose header and footer for an expected transactions file.

    Parameters
    ----------
    file_key
        key of `FILE_CONFIG` describing nature of file being composed.

    Returns
    -------
    tuple
        [0] header
        [1] footer
    """
    config = FILE_CONFIG[file_key]
    plugin, tag, comment = config["plugin"], config["tag"], config["comment"]

    header = f"""option "title" "{config['title']}"\n"""

    for k, v in get_account_root_names().items():
        if BC_DEFAULT_ACCOUNT_ROOT_NAMES[k] == v:
            continue
        header += f'option "{k}" "{v}"\n'

    if plugin is not None:
        header += f'plugin "{plugin}"\n'
    header += f"pushtag #{tag}\n"
    if comment is not None:
        for line in comment.split("\n"):
            header += f";; {line}\n"

    footer = ""
    if (post_comment := config["post_comment"]) is not None:
        for line in post_comment.split("\n"):
            footer += f";; {line}\n"
    footer += f"poptag #{tag}\n"
    return header, footer


def create_beanahead_file(
    file_key: str, dirpath: str | None = None, filename: str | None = None
):
    """Create a new beanahead file.

    Parameters
    ----------
    file_key
        key of `FILE_CONFIG` describing nature of file to be created. From:
            "x" - Expected Transactions Ledger
            "rx" - Regular Expected Transactions Ledger
            "rx_def" - Regular Expected Transaction Definitions file

    dirpath
        Path to directory in which to create new file. Can be absolute or
        relative to the current working directory. Must be an existing
        directory. By default, the current working directory.

    filename
        Name of new beanahead file. By default, same as `file_key`. If an
        extension is not included then the default extension will be added
        (as defined in the configuration file).

    Raises
    ------
    NotADirectoryError
        If `dirpath` does not represent a path to an existing directory.

    FileExistsError
        If .beancount file with `filename` already exists at `dirpath`.
    """
    file_key = file_key.lower()
    validate_file_key(file_key)

    dirpath = "." if dirpath is None else dirpath
    dirpath_ = Path(dirpath).absolute()
    if not dirpath_.is_dir():
        raise NotADirectoryError(dirpath_)

    filename = file_key if filename is None else filename
    path = dirpath_ / filename
    if not path.suffix:
        path = path.with_suffix(config.SETTINGS.extension)
    if path.is_file():
        raise FileExistsError(path)

    header, footer = compose_header_footer(file_key)
    content = header + "\n\n" + footer

    with path.open("wt", encoding=config.ENCODING) as file:
        file.write(content)


def verify_path(path: Path):
    """Verify path to an existing .beancount file.

    Parameters
    ----------
    path
        Path to be verified.

    Raises
    ------
    BeancountFileExistsError
        If path does not have .beancount suffix.
        If path does not represent an existing file.

    """
    if not path.is_file():
        raise BeancountFileExistsError(f"{path} is not a path to an existing file.")


def verify_files_key(path: Path, file_key: str):
    """Verify beancount file is a beanahead file of a specific nature.

    Parameters
    ----------
    path
        path representing beancount file to be verified.

    file_key
        key of `FILE_CONFIG` describing file nature to be verified.

    Raises
    ------
    BeancountFileKeyError
        If the first line of file at `path` does represent the expected
        title for `file_key`.

    See Also
    ---------
    get_verified_file_key : Verify path to a beanahead file and return its
        file key.
    get_verified_ledger_file_key : Verify path to a beanahead ledger file
        and return its file key.
    """
    with path.open("rt", encoding=config.ENCODING) as file:
        first_line = file.readline()
    title = FILE_CONFIG[file_key]["title"]
    if not first_line[:-2].endswith(title):
        raise BeanaheadFileKeyError(
            f"The first line of '{path}' does not contain the expected title '{title}'."
        )


def get_unverified_path(filepath: str) -> Path:
    """Return unverified path to a beancount file.

    Parameters
    ----------
    filepath
        Path to .beancount file, either absolute or relative to the cwd.
        It is not necessary to include the default extension (as defined
        in the configuration file, for example '.beancount').
    """
    path = Path(filepath).absolute()
    if not path.suffix:
        path = path.with_suffix(config.SETTINGS.extension)
    return path


def get_verified_path(filepath: str, file_key: str | None = None) -> Path:
    """Return verified path to a beancount file.

    Parameters
    ----------
    filepath
        Path to .beancount file, either absolute or relative to the cwd.
        It is not necessary to include the .beancount extension.

    file_key
        key of `FILE_CONFIG` describing nature of file to be verified.
        If passed then the .beancount file will additionally be verified
        as of the nature described by the file_key (raises
        BeanaheadFileKeyError if not verified).
    """
    path = get_unverified_path(filepath)
    verify_path(path)
    if file_key is not None:
        verify_files_key(path, file_key)
    return path


def get_options(path: Path) -> dict:
    """Get options for a beancount file.

    Parameters
    ----------
    path
        Path to beancount file. Path is NOT verified.

    Returns
    -------
    dict
        Options as mapping of 'option name' : value.
    """
    entries, errors, options = loader.load_file(path)
    return options


def set_account_root_names(filepath: str) -> dict[str, str]:
    """Set account root names from options defined on a beancount file.

    Parameters
    ----------
    filepath: str
        Path to .beancount file containing options defining account root
        names, for example:
            option "name_assets" "AltAssetsRootName"

        Can be passed as either absolute path or path relative to the cwd.
        It is not necessary to include the. beancount extension.
        For example, "ledger" would refer to the file 'ledger.beancount' in
        the cwd.

    Returns
    -------
    account_root_names: dict[str, str]
        Newly set account root names.
    """
    path = get_verified_path(filepath)
    options = get_options(path)
    names = {k: options[k] for k in BC_DEFAULT_ACCOUNT_ROOT_NAMES if k in options}
    return config.set_account_root_names(names)


def get_verified_file_key(path: Path) -> str:
    """Verify path to a beanahead file and return its file key.

    Parameters
    ----------
    path
        Path to file to be verified.

    Returns
        file_key, as key of `FILE_CONFIG`, corresponding with file at
        `path`.

    Raises
    ------
    BeancountFileExistsError
        If `path` does not represent a beancount file.

    BeanaheadFileExistsError
        If `path` does not represent a beanahead file.

    See Also
    ---------
    get_verified_ledger_file_key : Verify path to a beanahead ledger file
        and return its file key.
    verify_files_key : Verify beancount file is a beanahead file of a
        specific nature.
    """
    verify_path(path)
    opts = get_options(path)
    title = opts["title"]
    all_titles = []
    for file_key, config_ in FILE_CONFIG.items():
        if title == (config_title := config_["title"]):
            return file_key
        all_titles.append(config_title)
    titles_string = "\n".join(all_titles)
    raise BeanaheadFileExistsError(
        f"'{path}' does not represent a beanahead file. Beanahead files have"
        f" one of the following titles:\n{titles_string}"
    )


def get_verified_ledger_file_key(path: Path) -> str:
    """Verify path to a beanahead ledger file and return its file key.

    Parameters
    ----------
    path
        Path to a beanahead Expected Transactions Ledger file.

    Raises
    ------
    BeancountFileExistsError
        If `path` does not represent a beancount file.

    BeanaheadFileExistsError
        If `path` does not represent a beanahead file.

    BeanaheadLedgerFileExistsError
        If `path` represents a beanahead file, but not a beanahead
        Expected Transactions Ledger.

    See Also
    ---------
    get_verified_file_key : Verify path to a beanahead file and return its
        file key.
    verify_files_key : Verify beancount file is a beanahead file of a
        specific nature.
    """
    file_key = get_verified_file_key(path)
    if file_key not in LEDGER_FILE_KEYS:
        raise BeanaheadLedgerFileExistsError(path)
    return file_key


def get_unverified_entries(path: Path) -> data.Entries:
    """Get entries from a ledger file.

    Function is unconcerned with whether errors are raised when entries
    are loaded from the ledger.

    Parameters
    ----------
    path
        Path to ledger file.

    Returns
    -------
    data.Entries
        Entries extracted from unverified ledger file.
    """
    entries, errors, options = loader.load_file(path)
    return entries


def get_verified_entries(path: Path) -> data.Entries:
    """Get verified entries for a ledger file.

    Parameters
    ----------
    path
        Path to ledger file.

    Raises
    ------
    BeancountLoaderErrors
        If beancount raises any error when loading / attempting to
        load the ledger file.

    Returns
    -------
    data.Entries
        Entries extracted from verified ledger file.
    """
    entries, errors, options = loader.load_file(path)
    if errors:
        raise BeancountLoaderErrors(path, errors)
    return entries


def extract_txns(entries: data.Entries) -> list[Transaction]:
    """Return only entries that are transactions.

    Parameters
    ----------
    entries
        Entries to be filtered.
    """
    return [entry for entry in entries if isinstance(entry, Transaction)]


def get_unverified_txns(path: Path) -> list[Transaction]:
    """Get unverified transactions from a ledger file.

    Function is unconcerned with whether errors are raised when entries
    are loaded from the ledger.

    Parameters
    ----------
    path
        Path to ledger file.

    Returns
    -------
    list of Transaction
        Transactions extracted from unverified ledger file.
    """
    entries = get_unverified_entries(path)
    return extract_txns(entries)


def get_verified_txns(path: Path) -> list[Transaction]:
    """Get verified transactions from a ledger file.

    Parameters
    ----------
    path
        Path to ledger file.

    Returns
    -------
    list of Transaction
        Transactions extracted from verified ledger file.
    """
    entries = get_verified_entries(path)
    return extract_txns(entries)


def is_expired(txn: Transaction) -> bool:
    """Query if a transaction has expired.

    Returns
    -------
    bool
        True if `txn` is dated prior to today. False otherwise.
    """
    return txn.date < TODAY


def get_expired_txns(txns: list[Transaction]) -> list[Transaction]:
    """Get transactions dated prior to today."""
    return [txn for txn in txns if is_expired(txn)]


def remove_txns(
    txns: list[Transaction], txns_to_remove: list[Transaction]
) -> list[Transaction]:
    """Remove specific transactions from a list of transactions.

    All `txns_to_remove` must be present in `txns`.

    Parameters
    ----------
    txns
        List of transactions from which transactions are to be removed.

    txns_to_remove
        List of transactions to remove from `txns`.

    Returns
    -------
    Copy of `txns` excluding transactions in `txns_to_remove`.

    Raises
    ------
    ValueError
        If any item in txns_to_remove` is not in `txns`.
    """
    rtrn = [txn for txn in txns if txn not in txns_to_remove]
    if len(rtrn) > len(txns) - len(txns_to_remove):
        not_in_txns = [txn for txn in txns_to_remove if txn not in txns]
        assert not_in_txns
        raise ValueError(
            "The following items are in `txns_to_remove` although not"
            f" in `txns`:\n{not_in_txns}."
        )
    return rtrn


def reverse_automatic_balancing(txn: Transaction) -> Transaction:
    """Return copy of a transaction with any interpolation reversed.

    Parameters
    ----------
    txn
        Transaction from which to reverse any automatic balancing.
    """
    new_postings = []
    for posting in txn.postings:
        if AUTOMATIC_META in posting.meta:
            meta = {k: v for k, v in posting.meta.items() if k != AUTOMATIC_META}
            posting = posting._replace(units=None, meta=meta)
        new_postings.append(posting)
    return txn._replace(postings=new_postings)


def is_assets_account(string: str) -> bool:
    """Query if a string represents an assets account.

    Examples
    --------
    >>> is_assets_account("Expenses:Home:Electricity")
    False
    >>> is_assets_account("Equity:Opening-Balances")
    False
    >>> is_assets_account("Assets:US:BofA:Checking")
    True
    """
    return is_account_type(get_account_root_names()["name_assets"], string)


def is_balance_sheet_account(string: str) -> bool:
    """Query if a string represents a balance sheet account.

    Balance sheet accounts considered as "Assets" and "Liabilities" only.

    Parameters
    ----------
    txn
        Transaction to query.

    Examples
    --------
    >>> is_balance_sheet_account("Assets:US:BofA:Checking")
    True
    >>> is_balance_sheet_account("Liabilities:US:Chase:Slate")
    True
    >>> is_balance_sheet_account("Expenses:Home:Electricity")
    False
    >>> is_balance_sheet_account("Equity:Opening-Balances")
    False
    >>> is_balance_sheet_account("Income:US:BayBook:Match401k")
    False
    """
    account_root_names = get_account_root_names()
    return any(
        is_account_type(acc_type, string)
        for acc_type in [
            account_root_names["name_assets"],
            account_root_names["name_liabilities"],
        ]
    )


def get_balance_sheet_accounts(txn: Transaction) -> list[str]:
    """Return all balance sheet accounts associated with a transaction.

    Balance sheet accounts considered as "Assets" and "Liabilities" only.

    Parameters
    ----------
    txn
        Transaction to query.

    Returns
    -------
    list of str
        All balance sheet accounts to which postings are made.
    """
    accounts, seen = [], []
    for post in txn.postings:
        account = post.account
        if is_balance_sheet_account(account) and account not in seen:
            accounts.append(account)
        seen.append(account)
    return accounts


def get_content(path: Path) -> str:
    """Get all content from a file.

    Parameters
    ----------
    path
        Path to file from which to get all content.
    """
    with path.open("r", encoding=config.ENCODING) as file:
        content = file.read()
    return content


def clean_rx_meta(txn: Transaction) -> Transaction:
    """Clean a regular expected transaction's meta fields for printing.

    Removes meta fields that have default values.

    Parameters
    ----------
    txn
        Regular Expected Transaction.

    Returns
    -------
    Transaction
        Copy of `txn` with cleaned copy of `txn.meta`.
    """
    meta = copy.deepcopy(txn.meta)
    for k, v in RX_META_DFLTS.items():
        if meta[k] is v:
            del meta[k]
    return txn._replace(meta=meta)


def prepare_for_printer(txn: Transaction) -> Transaction:
    """Prepare a transaction for printing to a .beancount file.

        Reverses any interpolation of numbers for missing fields.

        Removes meta fields that have default values (for Regular Expected
        Transactions).

        Removes any expected transactions tags.

    Parameters
    ----------
    Transaction
        Transaction to prepare for printing.

    Returns
    -------
    Transaction
        Copy of `txn` suitable for printing to a ledger or definitions
        file.
    """
    txn = reverse_automatic_balancing(txn)
    if TAG_RX in txn.tags:
        txn = clean_rx_meta(txn)
    if TAGS_X & txn.tags:
        txn = remove_tags(txn, TAGS_X)
    return txn


def compose_entries_content(entries: data.Directive | data.Entries) -> str:
    """Return printable string representing one or more entries.

    Parameters
    ----------
    entries
        Entries to comprise content.
    """
    if not isinstance(entries, list):
        entries = [entries]
    content = ""
    for entry in entries:
        if isinstance(entry, Transaction):
            entry = prepare_for_printer(entry)
        content += printer.format_entry(entry) + "\n"
    return content[:-1]


def compose_new_content(file_key: str, txns_content: str) -> str:
    """Compose full content of an expected transactions .beancount file.

    Parameters
    ----------
    file_key
        key of `FILE_CONFIG` describing nature of file being composed.

    txns_content
        Content of transactions to be included to file.

    Returns
    -------
    str
        Content to overwrite an expected transactions .beancount file of
        nature described by `file_key`.

    Raises
    ------
    ValueError
        If created content parses with syntax errors.
    """
    header, footer = compose_header_footer(file_key)
    content = header + "\n\n" + txns_content + "\n\n" + footer
    _, errors, _ = parser.parse_string(content)
    if errors:
        raise ValueError(f"New content parses with following errors: {errors}")
    return content


def write(path: Path, content: str):
    """Write content to path."""
    with path.open("wt", encoding=config.ENCODING) as file:
        file.write(content)


def overwrite_file(path: Path, content: str):
    """Overwrite file with content for an expected transactions file.

    Parameters
    ----------
    path
        Path to file to be overwritten.

    content
        Content to write to file at `path`.

    Raises
    ------
    ValueError
        If `content` would subsequently load with errors.
    """
    _, errors, _ = parser.parse_string(content)
    if errors:
        raise ValueError(
            f"{path} has not been overwritten as content would parse with the"
            f" following errors: {errors}"
        )
    write(path, content)


def create_ledger_content(file_key: str, txns: list[Transaction]) -> str:
    """Create content for a beanahead ledger file.

    Parameters
    ----------
    file_key
        key of `FILE_CONFIG` describing nature of ledger for which
        content required.

    txns
        List of all expected transactions to be included in the content.
    """
    validate_ledger_file_key(file_key)
    txns_content = compose_entries_content(txns)
    return compose_new_content(file_key, txns_content)


def remove_txns_from_ledger(path: Path, txns: list[Transaction]):
    """Remove txns from an Expected Transactions Ledger.

    Expected Transactions Ledger at `path` will be overwritten with
    existing transactions excluding `txns`.

    NOTE Entries will be sorted and any comments on the the existing file
    will be lost.

    Parameters
    ----------
    path
        Path to Expected Transactions Ledger from which `txns` are to be
        removed.

    txns
        Transactions to remove from the ledger.
    """
    file_key = get_verified_ledger_file_key(path)
    existing_txns = get_unverified_txns(path)
    retained_txns = remove_txns(existing_txns, txns)
    new_content = create_ledger_content(file_key, retained_txns)
    overwrite_file(path, new_content)


def compile_strings_regex(
    strings: str | list[str],
    flags=re.I,  # noqa: E741
) -> re.Pattern:
    """Return regex to match any string of one or more strings.

    Will match if strings overlap.

    Parameters
    ----------
    strings
        Strings to match.

    flags : default: re.I
        re module flags. Combine more than one with | operator. Pass None
        to not add any flag.

    Returns
    -------
    re.Pattern
        Compiled regex.

    Examples
    --------
    >>> strings = ["one", "another"]
    >>> regex = compile_strings_regex(strings)
    >>> passes = [
    ...     "one for a match",  # at start of string
    ...     "and another",  # at end of string
    ...     "this one matches",  # in middle of string
    ...     "and another one",  # two match
    ...     "never alone you stand",  # not as a word
    ...     "oneanother",
    ...     "anotherone",
    ...     "aaonebb",
    ...     "aanotheroneb",
    ... ]
    >>> fails = [
    ...     "no match here",
    ...     "not on e anothe rmatch",
    ... ]
    >>> for pass_ in passes:
    ...     assert regex.match(pass_) is not None
    >>> for fail in fails:
    ...     assert regex.match(fail) is None
    """
    if isinstance(strings, str):
        strings = [strings]
    regex = "^.*(?=" + "|".join([rf"{string}" for string in strings]) + ").*$"
    return re.compile(regex, flags=flags)


def get_input(text: str) -> str:
    """Get user input.

    Parameters
    ----------
    text
        String to introduce request for user input.

    Returns
        User input

    Notes
    -----
    Function included to facilitate mocking user input when testing.
    """
    print_it(text, end=": ")
    return input()


def response_is_valid_number(response: str, max_value: int) -> bool:
    """Query if a response represents an integer less than a max_value.

    Examples
    --------
    >>> response_is_valid_number('2', 3)
    True
    >>> response_is_valid_number('3', 3)
    True
    >>> response_is_valid_number('4', 3)
    False
    >>> response_is_valid_number('0', 3)
    True
    >>> response_is_valid_number('-1', 3)
    False
    >>> response_is_valid_number('2.0', 3)
    False
    >>> response_is_valid_number('hello', 3)
    False
    """
    try:
        response = int(response)
    except ValueError:
        return False
    return 0 <= response <= max_value


def inject_txns(injection: str, ledger: str):
    """Inject new transactions to a ledger.

    New transactions are injected by way of appending the contents of
    the `injection` file to the end of the `ledger` file.

    Parameters
    ----------
    All parameters can be passed as either absolute paths or paths relative
    to the cwd. It is not necessary to include the. beancount extension.
    For example, "rx" would refer to the file 'rx.beancount' in the cwd.

    injection : str
        Path to .beancount file containing the new transactions to be
        injected to `ledger`.

    ledger: str
        Path to .beancount ledger file to which the new transactions are
        to be injected.
    """
    injection_path = get_verified_path(injection)
    content = get_content(injection_path)
    if content.startswith(HEADER):
        content = content[len(HEADER) :]
    content = "\n" + content
    ledger_path = get_verified_path(ledger)

    with ledger_path.open("at", encoding=config.ENCODING) as file:
        file.write(content)


def remove_tags(txn: Transaction, tags: str | list | set) -> Transaction:
    """Remove a specific tag or tags from a transaction.

    Parameters
    ----------
    txn
        Transaction from which to remove `tags`.

    tags
        Tags to remove from `txn`.

    Returns
    -------
    Transaction
        Copy of `txn` with `tags` removed.
        If `txn` has no tag of `tags` then returns `txn` as received.
    """
    tags = set([tags]) if isinstance(tags, str) else set(tags)
    new_tags = txn.tags - tags
    if new_tags != txn.tags:
        return txn._replace(tags=new_tags)
    return txn


def add_tags(txn: Transaction, tags: str | list | set) -> Transaction:
    """Add a specific tag or tags to a transaction.

    Parameters
    ----------
    txn
        Transaction to which to add `tags`.

    tags
        Tags to add to `txn`.

    Returns
    -------
    Transaction
        Copy of `txn` with `tags` added.
        If `txn` already has all `tags` then returns `txn` as received.
    """
    tags = set([tags]) if isinstance(tags, str) else set(tags)
    new_tags = txn.tags | tags
    if new_tags != txn.tags:
        return txn._replace(tags=new_tags)
    return txn
